from typing import Any, Optional
from essentials import json as JSON
from blacksheep import Request, Response
from blacksheep.utils import BytesOrStr
from blacksheep.server.routing import RoutesRegistry
from blacksheep.server.responses import (json,
                                         pretty_json,
                                         text,
                                         status_code,
                                         permanent_redirect,
                                         temporary_redirect,
                                         moved_permanently,
                                         see_other,
                                         redirect,
                                         not_found,
                                         no_content,
                                         not_modified,
                                         forbidden,
                                         unauthorized,
                                         bad_request,
                                         accepted,
                                         created,
                                         MessageType)


# singleton router used to store initial configuration, before the application starts
# this is used as *default* router for controllers, but it can be overridden - see for example tests in test_controllers
router = RoutesRegistry()


head = router.head
get = router.get
post = router.post
put = router.put
patch = router.patch
delete = router.delete
trace = router.trace
options = router.options
connect = router.connect


class ControllerMeta(type):

    def __init__(cls, name, bases, attr_dict):
        super().__init__(name, bases, attr_dict)

        for value in attr_dict.values():
            if hasattr(value, 'route_handler'):
                setattr(value, 'controller_type', cls)


class Controller(metaclass=ControllerMeta):
    """Base class for all controllers."""

    @classmethod
    def route(cls) -> Optional[str]:
        """
        The base route to be used by all request handlers defined on a controller type.
        Override this class method in subclasses, to implement base routes.
        """
        return None

    async def on_request(self, request: Request):
        """Extensibility point: controllers support executing a function at each web request.
        This method is executed before model binding happens: it is possible to read values from the request:
        headers are immediately available, and the user can decide to wait for the body (e.g. await request.json()).

        Since controllers support dependency injection, it is possible to apply logic using dependent services,
        specified in the __init__ constructor.
        """

    async def on_response(self, response: Response):
        """Callback called on every response returned using controller's methods."""

    def status_code(self, status: int = 200, message: MessageType = None) -> Response:
        """Returns a plain response with given status, with optional message; sent as plain text or JSON."""
        return status_code(status, message)

    def ok(self, message: MessageType = None) -> Response:
        """Returns an HTTP 200 OK response, with optional message; sent as plain text or JSON."""
        return status_code(200, message)

    def created(self, location: BytesOrStr, value: Any = None) -> Response:
        """Returns an HTTP 201 Created response, to the given location and with optional JSON content."""
        return created(location, value)

    def accepted(self, message: MessageType = None) -> Response:
        """Returns an HTTP 202 Accepted response, with optional message; sent as plain text or JSON."""
        return accepted(message)

    def no_content(self) -> Response:
        """Returns an HTTP 204 No Content response."""
        return no_content()

    def json(self, data, status: int = 200, dumps=JSON.dumps) -> Response:
        """Returns a response with application/json content, and given status (default HTTP 200 OK)."""
        return json(status, data, dumps)

    def pretty_json(self, data: Any, status: int = 200, dumps=JSON.dumps, indent: int = 4) -> Response:
        """Returns a response with indented application/json content, and given status (default HTTP 200 OK)."""
        return pretty_json(data, status, dumps, indent)

    def text(self, value: str, status: int = 200) -> Response:
        """Returns a response with text/plain content, and given status (default HTTP 200 OK)."""
        return text(value, status)

    def moved_permanently(self, location: BytesOrStr) -> Response:
        """Returns an HTTP 301 Moved Permanently response, to the given location"""
        return moved_permanently(location)

    def redirect(self, location: BytesOrStr) -> Response:
        """Returns an HTTP 302 Found response (commonly called redirect), to the given location"""
        return redirect(location)

    def see_other(self, location: BytesOrStr) -> Response:
        """Returns an HTTP 303 See Other response, to the given location."""
        return see_other(location)

    def not_modified(self) -> Response:
        """Returns an HTTP 304 Not Modified response."""
        return not_modified()

    def temporary_redirect(self, location: BytesOrStr) -> Response:
        """Returns an HTTP 307 Temporary Redirect response, to the given location."""
        return temporary_redirect(location)

    def permanent_redirect(self, location: BytesOrStr) -> Response:
        """Returns an HTTP 308 Permanent Redirect response, to the given location."""
        return permanent_redirect(location)

    def bad_request(self, message: MessageType = None) -> Response:
        """Returns an HTTP 400 Bad Request response, with optional message; sent as plain text or JSON."""
        return bad_request(message)

    def unauthorized(self, message: MessageType = None) -> Response:
        """Returns an HTTP 401 Unauthorized response, with optional message; sent as plain text or JSON."""
        return unauthorized(message)

    def forbidden(self, message: MessageType = None) -> Response:
        """Returns an HTTP 403 Forbidden response, with optional message; sent as plain text or JSON."""
        return forbidden(message)

    def not_found(self, message: MessageType = None) -> Response:
        """Returns an HTTP 404 Not Found response, with optional message; sent as plain text or JSON."""
        return not_found(message)



import re
from abc import abstractmethod
from functools import lru_cache
from blacksheep import HttpMethod
from collections import defaultdict
from urllib.parse import unquote
from typing import Callable, Dict, Optional, List, Any
from blacksheep.utils import ensure_bytes, ensure_str, BytesOrStr

__all__ = ['Router', 'Route', 'RouteMatch', 'RouteDuplicate', 'RegisteredRoute', 'RoutesRegistry']


_route_all_rx = re.compile(b'\\*')
_route_param_rx = re.compile(b'/:([^/]+)')
_named_group_rx = re.compile(b'\\?P<([^>]+)>')
_escaped_chars = {b'.', b'[', b']', b'(', b')'}


def _get_regex_for_pattern(pattern):

    for c in _escaped_chars:
        if c in pattern:
            pattern = pattern.replace(c, b'\\' + c)
    if b'*' in pattern:
        pattern = _route_all_rx.sub(br'(?P<tail>.+)', pattern)
    if b'/:' in pattern:
        pattern = _route_param_rx.sub(br'/(?P<\1>[^\/]+)', pattern)

    # NB: following code is just to throw user friendly errors;
    # regex would fail anyway, but with a more complex message 'sre_constants.error: redefinition of group name'
    # we only return param names as they are useful for other things
    param_names = []
    for p in _named_group_rx.finditer(pattern):
        param_name = p.group(1)
        if param_name in param_names:
            raise ValueError(f'cannot have multiple parameters with name: {param_name}')

        param_names.append(param_name)

    if len(pattern) > 1 and not pattern.endswith(b'*'):
        # NB: the /? at the end, ensures that a route is matched both with a trailing slash or not
        pattern = pattern + b'/?'
    return re.compile(b'^' + pattern + b'$', re.IGNORECASE), param_names


class RouteException(Exception):
    ...


class RouteDuplicate(RouteException):

    def __init__(self, method, pattern, current_handler):
        method = ensure_str(method)
        pattern = ensure_str(pattern)
        super().__init__(f'Cannot register route pattern `{pattern}` for `{method}` more than once. '
                         f'This pattern is already registered for handler {current_handler.__qualname__}.')
        self.method = method
        self.pattern = pattern
        self.current_handler = current_handler


class RouteMatch:

    __slots__ = ('values',
                 'handler')

    def __init__(self, route: 'Route', values: Optional[Dict[str, bytes]]):
        self.handler = route.handler
        self.values = {k: unquote(v.decode('utf8')) for k, v in values.items()} \
            if values else None  # type: Optional[Dict[str, str]]

    def __repr__(self):
        return f'<RouteMatch {id(self)}>'


class Route:

    __slots__ = ('handler',
                 'pattern',
                 'has_params',
                 'param_names',
                 '_rx')

    def __init__(self, pattern: bytes, handler: Any):
        if isinstance(pattern, str):
            pattern = pattern.encode('utf8')
        if pattern == b'':
            pattern = b'/'
        if len(pattern) > 1 and pattern.endswith(b'/'):
            pattern = pattern.rstrip(b'/')
        pattern = pattern.lower()
        self.handler = handler
        self.pattern = pattern
        self.has_params = b'*' in pattern or b':' in pattern
        rx, param_names = _get_regex_for_pattern(pattern)
        self._rx = rx
        self.param_names = [name.decode('utf8') for name in param_names]

    @property
    def full_pattern(self) -> bytes:
        return self._rx.pattern

    def match(self, value: bytes):
        if not self.has_params and value.lower() == self.pattern:
            return RouteMatch(self, None)

        match = self._rx.match(value)

        if not match:
            return None

        return RouteMatch(self, match.groupdict() if self.has_params else None)

    def __repr__(self):
        return f'<Route {self.pattern}>'


class RouterBase:

    def mark_handler(self, handler: Callable):
        setattr(handler, 'route_handler', True)

    @abstractmethod
    def add(self, method: str, pattern: BytesOrStr, handler: Callable):
        ...

    def add_head(self, pattern, handler):
        self.add(HttpMethod.HEAD, pattern, handler)

    def add_get(self, pattern, handler):
        self.add(HttpMethod.GET, pattern, handler)

    def add_post(self, pattern, handler):
        self.add(HttpMethod.POST, pattern, handler)

    def add_put(self, pattern, handler):
        self.add(HttpMethod.PUT, pattern, handler)

    def add_delete(self, pattern, handler):
        self.add(HttpMethod.DELETE, pattern, handler)

    def add_trace(self, pattern, handler):
        self.add(HttpMethod.TRACE, pattern, handler)

    def add_options(self, pattern, handler):
        self.add(HttpMethod.OPTIONS, pattern, handler)

    def add_connect(self, pattern, handler):
        self.add(HttpMethod.CONNECT, pattern, handler)

    def add_patch(self, pattern, handler):
        self.add(HttpMethod.PATCH, pattern, handler)

    def add_any(self, pattern, handler):
        self.add('*', pattern, handler)

    def head(self, pattern='/'):
        def decorator(f):
            self.add(HttpMethod.HEAD, pattern, f)
            return f
        return decorator

    def get(self, pattern='/'):
        def decorator(f):
            self.add(HttpMethod.GET, pattern, f)
            return f
        return decorator

    def post(self, pattern='/'):
        def decorator(f):
            self.add(HttpMethod.POST, pattern, f)
            return f
        return decorator

    def put(self, pattern='/'):
        def decorator(f):
            self.add(HttpMethod.PUT, pattern, f)
            return f
        return decorator

    def delete(self, pattern='/'):
        def decorator(f):
            self.add(HttpMethod.DELETE, pattern, f)
            return f
        return decorator

    def trace(self, pattern='/'):
        def decorator(f):
            self.add(HttpMethod.TRACE, pattern, f)
            return f
        return decorator

    def options(self, pattern='/'):
        def decorator(f):
            self.add(HttpMethod.OPTIONS, pattern, f)
            return f
        return decorator

    def connect(self, pattern='/'):
        def decorator(f):
            self.add(HttpMethod.CONNECT, pattern, f)
            return f
        return decorator

    def patch(self, pattern='/'):
        def decorator(f):
            self.add(HttpMethod.PATCH, pattern, f)
            return f
        return decorator


class Router(RouterBase):

    __slots__ = ('routes',
                 '_map',
                 '_fallback')

    def __init__(self):
        self._map = {}
        self._fallback = None
        self.routes = defaultdict(list)

    @property
    def fallback(self):
        return self._fallback

    @fallback.setter
    def fallback(self, value):
        if not isinstance(value, Route):
            if callable(value):
                self._fallback = Route(b'*', value)
                return
            raise ValueError('fallback must be a Route')
        self._fallback = value

    def __iter__(self):
        for key, routes in self.routes.items():
            for route in routes:
                yield route
        if self._fallback:
            yield self._fallback

    def _is_route_configured(self, method: bytes, route: Route):
        method_patterns = self._map.get(method)
        if not method_patterns:
            return False
        if method_patterns.get(route.full_pattern):
            return True
        return False

    def _set_configured_route(self, method: bytes, route: Route):
        method_patterns = self._map.get(method)
        if not method_patterns:
            self._map[method] = {route.full_pattern: route}
        else:
            method_patterns[route.full_pattern] = route

    def _check_duplicate(self, method: bytes, new_route: Route):
        if self._is_route_configured(method, new_route):
            current_route = self._map.get(method).get(new_route.full_pattern)
            raise RouteDuplicate(method, new_route.pattern, current_route.handler)
        self._set_configured_route(method, new_route)

    def add(self, method: str, pattern: BytesOrStr, handler: Any):
        self.mark_handler(handler)
        method = ensure_bytes(method)
        pattern = ensure_bytes(pattern)
        new_route = Route(pattern, handler)
        self._check_duplicate(method, new_route)
        self.add_route(method, new_route)

    def add_route(self, method: BytesOrStr, route: Route):
        self.routes[ensure_bytes(method)].append(route)

    @lru_cache(maxsize=1200)
    def get_match(self, method, value):
        method = ensure_bytes(method)
        value = ensure_bytes(value)

        for route in self.routes[method]:
            match = route.match(value)
            if match:
                return match
        return RouteMatch(self._fallback, None) if self.fallback else None


class RegisteredRoute:

    __slots__ = ('method', 'pattern', 'handler')

    def __init__(self, method: str, pattern: bytes, handler: Callable):
        self.method = method
        self.pattern = pattern
        self.handler = handler

    def __repr__(self):
        try:
            return f'<{self.__class__.__name__} {self.method} {self.pattern.decode()} {self.handler.__name__}>'
        except AttributeError:
            return f'<{self.__class__.__name__} at {id(self)}>'


class RoutesRegistry(RouterBase):
    """A registry for routes: not a full router able to get matches.
    Unlike a router, a registry does not throw for duplicated routes; because such routes can be modified
    when applied to an actual router.

    This class is meant to enable scenarios like base pattern for controllers.
    """

    def __init__(self):
        self.routes = []  # type: List[RegisteredRoute]

    def __iter__(self):
        yield from self.routes

    def add(self, method: str, pattern: BytesOrStr, handler: Callable):
        self.mark_handler(handler)
        self.routes.append(RegisteredRoute(method, pattern, handler))





def copy_special_attributes(source_method, wrapper):
    for name in {'auth',
                 'auth_policy',
                 'auth_schemes',
                 'allow_anonymous',
                 'controller_type',
                 'route_handler'}:
        if hasattr(source_method, name):
            setattr(wrapper, name, getattr(source_method, name))



from typing import Tuple, Optional, Generator
from blacksheep.contents import FormPart


def get_boundary(value: bytes):
    return value[:value.index(b'\n')+1]


def get_boundary_from_header(value: bytes):
    return value.split(b'=', 1)[1].split(b' ', 1)[0]


def _remove_last_crlf(value: bytes):
    if value.endswith(b'\r\n'):
        return value[:-2]
    if value.endswith(b'\n'):
        return value[:-1]
    return value


def split_multipart(value: bytes):
    """Splits a whole multipart/form-data payload into single parts without boundary."""
    value = value.strip(b' ')
    boundary = value[:value.index(b'\n')].rstrip(b'\r')

    for part in value.split(boundary):
        part = _remove_last_crlf(part.lstrip(b'\r\n'))
        if part == b'' or part == b'--':
            continue
        yield part


def split_headers(value: bytes):
    """Splits a whole portion of multipart form data representing headers into name, value pairs"""
    #
    # Examples of whole portions:
    #
    # Content-Disposition: form-data; name="two"
    # Content-Disposition: form-data; name="file_example"; filename="example-001.png"\r\nContent-Type: image/png
    #
    for raw_header in value.split(b'\r\n'):
        header_name, header_value = raw_header.split(b':', 1)
        yield header_name.lower(), header_value.lstrip(b' ')


def split_content_disposition_values(value: bytes) -> Tuple[bytes, Optional[bytes]]:
    """Parses a single header into key, value pairs"""
    for part in value.split(b';'):
        if b'=' in part:
            name, value = part.split(b'=', 1)
            yield name.lower().strip(b' '), value.strip(b'" ')
        else:
            yield b'type', part


def parse_content_disposition_values(value: bytes):
    return dict(split_content_disposition_values(value))


class CharsetPart(Exception):

    def __init__(self, default_charset: bytes):
        self.default_charset = default_charset


def parse_part(value: bytes, default_charset: Optional[bytes]) -> FormPart:
    """Parses a single multipart/form-data part."""
    raw_headers, data = value.split(b'\r\n\r\n', 1)

    headers = dict(split_headers(raw_headers))

    content_disposition = headers.get(b'content-disposition')

    if not content_disposition:
        # what to do? raise makes sense
        raise Exception('Missing Content-Disposition header in multipart/form-data part.')

    content_disposition_values = parse_content_disposition_values(content_disposition)

    field_name = content_disposition_values.get(b'name')

    if field_name == b'_charset_':
        # NB: handling charset...
        # https://tools.ietf.org/html/rfc7578#section-4.6
        raise CharsetPart(data)

    content_type = headers.get(b'content-type', None)  # TODO: b'text/plain' default to text plain?

    return FormPart(field_name,
                    data,
                    content_type,
                    content_disposition_values.get(b'filename'),
                    default_charset)


def parse_multipart(value: bytes) -> Generator[FormPart, None, None]:
    default_charset = None

    for part_bytes in split_multipart(value):

        try:
            yield parse_part(part_bytes, default_charset)
        except CharsetPart as charset:
            default_charset = charset.default_charset



import os
import aiofiles
from datetime import datetime
from blacksheep import Response, StreamedContent
from blacksheep.server.pathsutils import get_mime_type


def get_default_extensions():
    return {
        '.txt',
        '.css',
        '.js',
        '.jpeg',
        '.jpg',
        '.html',
        '.ico',
        '.png',
        '.woff',
        '.woff2',
        '.ttf',
        '.eot',
        '.svg',
        '.mp4',
        '.mp3'
    }


def unix_timestamp_to_datetime(ts):
    return datetime.utcfromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S').encode()


async def get_file_chunks(file_path: str, size_limit: int = 1024*64):
    async with aiofiles.open(file_path, mode='rb') as f:
        while True:
            chunk = await f.read(size_limit)

            if not chunk:
                break

            yield chunk


def get_file_data(file_path, file_size, size_limit=1024*64):
    # NB: if the file size is small, we read its bytes and return them;
    # otherwise, a lazy reader is returned; that returns the file in chunks
    if file_size > size_limit:
        async def file_chunker():
            async for chunk in get_file_chunks(file_path, size_limit):
                yield chunk
            yield b''

        return file_chunker

    async def file_getter():
        async with aiofiles.open(file_path, mode='rb') as file:
            data = await file.read()
            yield data
            yield b''
    return file_getter


def get_response_for_file(request, resource_path: str, cache_time: int):
    stat = os.stat(resource_path)
    file_size = stat.st_size
    modified_time = stat.st_mtime
    current_etag = str(modified_time).encode()
    previous_etag = request.if_none_match

    headers = [
        (b'Last-Modified', unix_timestamp_to_datetime(modified_time)),
        (b'ETag', current_etag)
    ]

    if cache_time > 0:
        headers.append((b'Cache-Control', b'max-age=' + str(cache_time).encode()))

    if previous_etag and current_etag == previous_etag:
        return Response(304, headers, None)

    if request.method == 'HEAD':
        headers.append((b'Content-Type', get_mime_type(resource_path)))
        headers.append((b'Content-Length', str(file_size).encode()))
        return Response(200, headers, None)

    return Response(200, 
                    headers,
                    StreamedContent(get_mime_type(resource_path),
                                    get_file_data(resource_path, file_size)))




class InvalidOperation(Exception):

    def __init__(self, message: str, inner_exception: object=None):
        super().__init__(message)
        self.inner_exception = inner_exception


class BadRequestFormat(Exception):

    def __init__(self, message: str, inner_exception: object=None):
        super().__init__(message)
        self.inner_exception = inner_exception


class HttpException(Exception):

    def __init__(self, status: int, message: str = 'HTTP Exception'):
        super().__init__(message)
        self.status = status


class BadRequest(HttpException):

    def __init__(self, message: str):
        super().__init__(400, message)


class NotFound(HttpException):

    def __init__(self):
        super().__init__(404)


class InvalidArgument(Exception):

    def __init__(self, message: str):
        super().__init__(message)


class MessageAborted(Exception):
    def __init__(self):
        super().__init__('The message was aborted before the client sent its whole content.')



from datetime import datetime
from typing import Optional, Union


def datetime_to_cookie_format(value: datetime) -> bytes: ...


def datetime_from_cookie_format(value: bytes) -> datetime: ...


class Cookie:

    def __init__(self,
                 name: bytes,
                 value: bytes,
                 expires: Optional[bytes] = None,
                 domain: Optional[bytes] = None,
                 path: Optional[bytes] = None,
                 http_only: bool = False,
                 secure: bool = False,
                 max_age: Optional[bytes] = None,
                 same_site: Optional[bytes] = None):
        self.name = name
        self.value = value
        self.expires = expires
        self._expiration = None
        self.domain = domain
        self.path = path
        self.http_only = http_only
        self.secure = secure
        self.max_age = max_age
        self.same_site = same_site

    def clone(self) -> Cookie: ...

    @property
    def expiration(self) -> datetime: ...

    @expiration.setter
    def expiration(self, value: datetime): ...

    def set_max_age(self, max_age: int): ...

    def __eq__(self, other: Union[str, bytes, Cookie]) -> bool: ...

    def __repr__(self):
        return f'<Cookie {self.name.decode()}: {self.value.decode()}>'


def parse_cookie(value: bytes) -> Cookie: ...




__author__ = 'Roberto Prevato <roberto.prevato@gmail.com>'

from .url import URL, InvalidURL
from .headers import Header, Headers
from .exceptions import HttpException
from .contents import (Content,
                       StreamedContent,
                       JsonContent,
                       FormContent,
                       FormPart,
                       TextContent,
                       HtmlContent,
                       MultiPartFormData,
                       parse_www_form)
from .cookies import (Cookie,
                      datetime_from_cookie_format,
                      datetime_to_cookie_format,
                      parse_cookie)
from .messages import Request, Response


class HttpMethod:
    GET = 'GET'
    HEAD = 'HEAD'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    TRACE = 'TRACE'
    OPTIONS = 'OPTIONS'
    CONNECT = 'CONNECT'
    PATCH = 'PATCH'



import json
import uuid
from typing import AsyncGenerator, Any, Callable, Optional, List, Dict, Union, Tuple


class Content:

    def __init__(self,
                 content_type: bytes,
                 data: bytes):
        self.type = content_type
        self.body = data
        self.length = len(data)

    async def read(self) -> bytes:
        return self.body


class StreamedContent(Content):

    def __init__(self,
                 content_type: bytes,
                 data_provider: AsyncGenerator[bytes]):
        self.type = content_type
        self.body = None
        self.length = -1
        self.generator = data_provider

    async def get_parts(self) -> AsyncGenerator[bytes]: ...


class ASGIContent(Content):

    def __init__(self, receive: Callable[[], bytes]):
        self.type = None
        self.body = None
        self.length = -1
        self.receive = receive

    def dispose(self): ...

    async def stream(self) -> AsyncGenerator[bytes]: ...

    async def read(self) -> bytes: ...


class TextContent(Content):

    def __init__(self, text: str):
        super().__init__(b'text/plain; charset=utf-8', text.encode('utf8'))


class HtmlContent(Content):

    def __init__(self, html: str):
        super().__init__(b'text/html; charset=utf-8', html.encode('utf8'))


class JsonContent(Content):

    def __init__(self, data: object, dumps: Callable[[Any], str] = json.dumps):
        super().__init__(b'application/json', dumps(data).encode('utf8'))


class FormContent(Content):

    def __init__(self, data: Union[Dict[str, str], List[Tuple[str, str]]]):
        """
        Creates a new instance of content with application/x-www-form-urlencoded type,
        and bytes data serialized from the given dictionary.

        :param data: data to be serialized.
        """
        super().__init__(b'application/x-www-form-urlencoded', ...)


class FormPart:

    def __init__(self,
                 name: bytes,
                 data: bytes,
                 content_type: Optional[bytes]=None,
                 file_name: Optional[bytes]=None,
                 charset: Optional[bytes] = None):
        self.name = name
        self.data = data
        self.file_name = file_name
        self.content_type = content_type
        self.charset = charset

    def __repr__(self):
        return f'<FormPart {self.name} - at {id(self)}>'


class MultiPartFormData(Content):

    def __init__(self, parts: List[FormPart]):
        self.parts = parts
        self.boundary = b'------' + str(uuid.uuid4()).replace('-', '').encode()
        super().__init__(b'multipart/form-data; boundary=' + self.boundary, ...)


def parse_www_form(content: str) -> Dict[str, Union[str, List[str]]]:
    """Parses application/x-www-form-urlencoded content"""


def write_www_form_urlencoded(data: Union[Dict[str, str], List[Tuple[str, str]]]) -> bytes: ...




from .session import (ClientSession,
                      CircularRedirectError,
                      MaximumRedirectsExceededError,
                      MissingLocationForRedirect,
                      ConnectionTimeout,
                      RequestTimeout)



from typing import Optional, Sequence
from blacksheep import Response, TextContent
from guardpost.synchronous.authorization import Requirement, UnauthorizedError
from guardpost.asynchronous.authorization import AuthorizationStrategy, AsyncRequirement, Policy


__all__ = ('auth',
           'AuthorizationStrategy',
           'AuthorizationWithoutAuthenticationError',
           'allow_anonymous',
           'get_authorization_middleware',
           'Requirement',
           'AsyncRequirement',
           'handle_unauthorized',
           'Policy')


def auth(policy: Optional[str] = None, *, authentication_schemes: Optional[Sequence[str]] = None):
    """Configures authorization for a decorated request handler, optionally with a policy.
    If no policy is specified, the default policy of the configured AuthorizationStrategy is used.

    :param policy: optional, name of the policy to use to achieve authorization.
    :param authentication_schemes: optional, authentication schemes to use for this handler. If not specified, all
    configured authentication handlers are used.
    """
    def decorator(f):
        f.auth = True
        f.auth_policy = policy
        f.auth_schemes = authentication_schemes
        return f
    return decorator


def allow_anonymous():
    """Configures a decorated request handler, to be usable for all users: anonymous and authenticated users."""
    def decorator(f):
        f.allow_anonymous = True
        return f
    return decorator


def get_authorization_middleware(strategy: AuthorizationStrategy):
    async def authorization_middleware(request, handler):
        identity = request.identity

        if hasattr(handler, 'allow_anonymous'):
            return await handler(request)

        if hasattr(handler, 'auth'):
            # function decorated by auth;
            await strategy.authorize(handler.auth_policy, identity)
        else:
            # function not decorated by auth; use the default policy
            await strategy.authorize(None, identity)

        return await handler(request)
    return authorization_middleware


class AuthorizationWithoutAuthenticationError(RuntimeError):

    def __init__(self):
        super().__init__('Cannot use an authorization strategy without an authentication strategy. '
                         'Use `use_authentication` method to configure it.')


def _get_www_authenticated_header_value_from_generic_unauthorized_error(error: UnauthorizedError):
    parts = [error.scheme]
    if error.error:
        parts.append(', error="{}"')


def get_www_authenticated_header_from_generic_unauthorized_error(error):
    if not error.scheme:
        return None

    return b'WWW-Authenticate', error.scheme.decode()


async def handle_unauthorized(app, request, http_exception: UnauthorizedError):
    www_authenticate = get_www_authenticated_header_from_generic_unauthorized_error(http_exception)
    return Response(401,
                    [www_authenticate] if www_authenticate else None,
                    content=TextContent('Unauthorized'))



import ssl
import asyncio
import httptools
import certifi
import weakref
from blacksheep import Request, Response, Content
from blacksheep.scribe import (is_small_request,
                               write_small_request,
                               write_request_without_body,
                               write_request,
                               request_has_body,
                               write_request_body_only)
from httptools import HttpParserError, HttpParserCallbackError
SECURE_SSLCONTEXT = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=certifi.where())
SECURE_SSLCONTEXT.check_hostname = True

INSECURE_SSLCONTEXT = ssl.SSLContext()
INSECURE_SSLCONTEXT.check_hostname = False


class IncomingContent(Content):

    def __init__(self, content_type: bytes):
        super().__init__(content_type, b'')
        self._body = bytearray()
        self._chunk = asyncio.Event()
        self.complete = asyncio.Event()

    def extend_body(self, chunk: bytes):
        self._body.extend(chunk)
        self._chunk.set()

    async def stream(self):
        while True:
            await self._chunk.wait()
            self._chunk.clear()

            yield bytes(self._body)
            self._body.clear()

            if self.complete.is_set():
                break

    async def read(self):
        await self.complete.wait()
        return bytes(self._body)


class ConnectionClosedError(Exception):

    def __init__(self, can_retry: bool):
        super().__init__('The connection was closed by the remote server.')
        self.can_retry = can_retry


class InvalidResponseFromServer(Exception):

    def __init__(self, inner_exception, message=None):
        super().__init__(message or 'The remote endpoint returned an invalid HTTP response.')
        self.inner_exception = inner_exception


class UpgradeResponse:

    def __init__(self, response, transport):
        self.response = response
        self.transport = transport


class ClientConnection(asyncio.Protocol):

    __slots__ = (
        'loop',
        'pool',
        'transport',
        'open',
        '_connection_lost',
        'writing_paused',
        'writable',
        'ready',
        'response_ready',
        'request',
        'expect_100_continue',
        '_pending_task',
        '_can_release',
        '_upgraded'
    )

    def __init__(self, loop, pool):
        self.loop = loop
        self.pool = weakref.ref(pool)
        self.transport = None
        self.open = False
        self.writing_paused = False
        self.writable = asyncio.Event()
        self.ready = asyncio.Event()
        self.response_ready = asyncio.Event()
        self.expect_100_continue = False
        self.request = None
        self.request_timeout = 20
        self.headers = []
        self.response = None
        self.parser = httptools.HttpResponseParser(self)
        self._connection_lost = False
        self._pending_task = None
        self._can_release = False
        self._upgraded = False

    def reset(self):
        self.headers = []
        self.request = None
        self.response = None
        self.writing_paused = False
        self.writable.set()
        self.expect_100_continue = False
        self.parser = httptools.HttpResponseParser(self)
        self._connection_lost = False
        self._pending_task = None
        self._can_release = False
        self._upgraded = False

    def pause_writing(self):
        super().pause_writing()
        self.writing_paused = True
        self.writable.clear()

    def resume_writing(self):
        super().resume_writing()
        self.writing_paused = False
        self.writable.set()

    def connection_made(self, transport):
        self.transport = transport
        self.open = True
        self.ready.set()

    async def _wait_response(self):
        await self.response_ready.wait()

        self._pending_task = False
        if self._can_release:
            self.loop.call_soon(self.release)

        response = self.response

        if 99 < response.status < 200:
            # Handle 1xx informational
            #  https://tools.ietf.org/html/rfc7231#section-6.2

            if response.status == 101:
                # 101 Upgrade is a final response as it's used to switch protocols with WebSockets handshake.
                # returns the response object with status 101 and access to transport
                self._upgraded = True
                return UpgradeResponse(response, self.transport)

            if response.status == 100 and self.expect_100_continue:
                await self._send_body(self.request)

            # ignore;
            self.response_ready.clear()
            self.headers = []

            # await the final response
            return await self._wait_response()

        if self._connection_lost:  # TODO: should also check if the response is not complete?
            raise ConnectionClosedError(False)

        self.response_ready.clear()
        return response

    async def _write_chunks(self, request, method):
        async for chunk in method(request):
            if self._can_release:
                # the server returned a response before we ended sending the request
                return await self._wait_response()

            if not self.open:
                raise ConnectionClosedError(False)

            if self.writing_paused:
                await self.writable.wait()
            self.transport.write(chunk)

    async def _send_body(self, request):
        await self._write_chunks(request, write_request_body_only)

    async def send(self, request):
        if not self.open:
            # NB: if the connection is closed here, it is always possible to try again with a new connection
            # instead, if it happens later; we cannot retry because we started sending a request
            raise ConnectionClosedError(True)

        self.request = request
        self._pending_task = True

        if request_has_body(request) and request.expect_100_continue():
            # don't send the body immediately; instead, wait for HTTP 100 Continue interim response from server
            self.expect_100_continue = True
            self.transport.write(write_request_without_body(request))

            return await self._wait_response()

        if is_small_request(request):
            self.transport.write(write_small_request(request))
        else:
            await self._write_chunks(request, write_request)

        return await self._wait_response()

    def close(self):
        if self.open:
            self.open = False
            if self.transport:
                self.transport.close()

    def data_received(self, data):
        try:
            self.parser.feed_data(data)
        except HttpParserCallbackError:
            self.close()
            raise
        except HttpParserError as pex:
            raise InvalidResponseFromServer(pex)

    def connection_lost(self, exc):
        self._connection_lost = True
        self.ready.clear()
        self.open = False

        if self._pending_task:
            self.response_ready.set()

    def on_header(self, name, value):
        self.headers.append((name, value))

    def on_headers_complete(self):
        status = self.parser.get_status_code()
        self.response = Response(
            status,
            self.headers,
            None
        )
        # NB: check if headers declare a content-length
        if self._has_content():
            self.response.content = IncomingContent(self.response.get_single_header(b'content-type'))

        self.response_ready.set()

    def _has_content(self):
        content_length = self.response.get_first_header(b'content-length')

        if content_length:
            try:
                content_length_value = int(content_length)
            except ValueError as value_error:
                # server returned an invalid content-length value
                raise InvalidResponseFromServer(value_error, f'The server returned an invalid value for'
                                                             f'the Content-Lenght header; value: {content_length}')
            return content_length_value > 0

        transfer_encoding = self.response.get_first_header(b'transfer-encoding')

        if transfer_encoding and b'chunked' in transfer_encoding:
            return True

        return False

    def on_message_complete(self):
        if self.response and self.response.content:
            self.response.content.complete.set()

        if self._pending_task:
            # the server returned a response before we ended sending the request,
            # the connection cannot be released now - this can happen for our Bad Requests
            self._can_release = True
            self.response_ready.set()
        else:
            # request-response cycle completed now,
            # the connection can be returned to its pool
            self.loop.call_soon(self.release)

    def release(self):
        if not self.open or self._upgraded:
            # if the connection was upgraded, its transport is used for web sockets,
            # it cannot return to its pool for other cycles
            return

        if self.parser.should_keep_alive():
            self.reset()
            pool = self.pool()

            if pool:
                pool.try_return_connection(self)
        else:
            self.close()

    def on_body(self, value: bytes):
        self.response.content.extend_body(value)



from blacksheep.messages import Request
from blacksheep.contents import Content


def get_status_line(status: int) -> bytes: ...

def is_small_request(request: Request) -> bool: ...

def request_has_body(request: Request) -> bool: ...

def write_small_request(request: Request) -> bytes: ...

def write_request_without_body(request: Request) -> bytes: ...

# TODO: how to annotate an async generator?
async def write_chunks(content: Content): ...



import os
import html
from pathlib import Path
from blacksheep.exceptions import InvalidArgument
from blacksheep.server.routing import Route
from blacksheep import Response, HtmlContent
from blacksheep.server.pathsutils import get_file_extension_from_name
from blacksheep.server.resources import get_resource_file_content
from blacksheep.exceptions import NotFound
from urllib.parse import unquote
from . import get_default_extensions, get_response_for_file


def get_files_to_serve(source_folder, extensions=None, recurse=False, root_folder=None):
    if extensions is None:
        extensions = get_default_extensions()

    if not root_folder:
        root_folder = source_folder

    p = Path(source_folder)

    if not p.exists():
        raise InvalidArgument('given root path does not exist')

    if not p.is_dir():
        raise InvalidArgument('given root path is not a directory')

    names = os.listdir(p)
    names.sort()
    dirs, nondirs = [], []

    for name in names:
        full_path = Path(os.path.join(source_folder, name))
        if os.path.isdir(full_path):
            dirs.append(full_path)
        else:
            nondirs.append(full_path)

    items = dirs + nondirs

    for item in items:
        item_path = str(item)

        if os.path.islink(item_path):
            continue

        if item.is_dir():
            if not recurse:
                yield {
                    'rel_path': item_path[len(root_folder)+1:],
                    'full_path': item_path,
                    'is_dir': True
                }
            else:
                for v in get_files_to_serve(Path(item_path),
                                            extensions,
                                            recurse,
                                            root_folder):
                    yield v
        else:
            extension = get_file_extension_from_name(item_path)

            if extension in extensions:
                yield {
                    'rel_path': item_path[len(root_folder)+1:],
                    'full_path': item_path,
                    'is_dir': False
                }


def get_files_list_html_response(template, parent_folder_path, contents):
    info = ''
    for item in contents:
        full_rel_path = html.escape(os.path.join(parent_folder_path, item.get("rel_path")))
        info += f'<li><a href="/{full_rel_path}">{full_rel_path}</a></li>'

    p = []
    whole_p = []
    for fragment in parent_folder_path.split('/'):
        if fragment:
            whole_p.append(html.escape(fragment))
            fragment_path = '/'.join(whole_p)
            p.append(f'<a href="/{fragment_path}">{html.escape(fragment)}</a>')

    # TODO: use chunked encoding here with HTML response
    return Response(200, content=HtmlContent(template.format_map({'path': '/'.join(p), 'info': info})))


def get_files_handler(source_folder_name, source_folder_full_path, discovery, cache_time, extensions):
    files_list_template = get_resource_file_content('fileslist.html') if discovery else None
    source_folder_full_path = source_folder_full_path.lower()

    async def file_getter(request):
        tail = unquote(request.route_values.get('tail')).lstrip('/')

        resource_path = os.path.join(source_folder_name, tail)

        if '../' in tail:
            abs_path = os.path.abspath(resource_path)
            if not str(abs_path).lower().startswith(source_folder_full_path):
                # someone is trying to read out of the static folder
                raise NotFound()

        if not os.path.exists(resource_path) or os.path.islink(resource_path):
            raise NotFound()

        if os.path.isdir(resource_path):
            if discovery:
                return get_files_list_html_response(files_list_template,
                                                    tail.rstrip('/'),
                                                    list(get_files_to_serve(resource_path.rstrip('/'))))
            else:
                raise NotFound()

        file_extension = get_file_extension_from_name(resource_path)

        if file_extension not in extensions:
            raise NotFound()

        return get_response_for_file(request, resource_path, cache_time)
    return file_getter


def serve_files(router,
                source_folder,
                extensions=None,
                discovery=False,
                cache_time=10800):
    if extensions is None:
        extensions = get_default_extensions()

    source_folder_path = Path(source_folder)

    if not source_folder_path.exists():
        raise InvalidArgument('given root path does not exist')

    if not source_folder_path.is_dir():
        raise InvalidArgument('given root path is not a directory')

    route = Route(b'*', get_files_handler(source_folder,
                                          os.path.abspath(source_folder),
                                          discovery,
                                          cache_time,
                                          extensions))
    router.add_route('GET', route)
    router.add_route('HEAD', route)



import json
from typing import Dict, Generator, List, Union, Optional, Any, Callable
from .url import URL
from .contents import Content, FormPart
from .headers import HeaderType, Headers
from .cookies import Cookie
from guardpost.authentication import Identity, User


class Message:

    @property
    def headers(self) -> Headers: ...

    def content_type(self) -> bytes: ...

    def with_content(self, content: Content) -> Message: ...

    def get_first_header(self, key: bytes) -> bytes: ...

    def get_headers(self, key: bytes) -> List[bytes]: ...

    def get_single_header(self, key: bytes) -> bytes: ...

    def remove_header(self, key: bytes): ...

    def has_header(self, key: bytes) -> bool: ...

    def add_header(self, name: bytes, value: bytes): ...

    def set_header(self, key: bytes, value: bytes): ...

    def content_type(self) -> bytes: ...

    async def read(self) -> Optional[bytes]: ...

    async def stream(self) -> Generator[bytes, None, None]: ...

    async def text(self) -> str: ...

    async def form(self) -> Union[Dict[str, str], Dict[str, List[str]], None]:
        """Returns values read either from multipart or www-form-urlencoded payload.

        This function adopts some compromises to provide a consistent api, returning a dictionary of key: values.
        If a key is unique, the value is a single string; if a key is duplicated (licit in both form types),
        the value is a list of strings.

        Multipart form parts values that can be decoded as UTF8 are decoded, otherwise kept as raw bytes.
        In case of ambiguity, use the dedicated `multiparts()` method.
        """

    async def multipart(self) -> List[FormPart]:
        """Returns parts read from multipart/form-data, if present, otherwise None"""

    def declares_content_type(self, type: bytes) -> bool: ...

    def declares_json(self) -> bool: ...

    def declares_xml(self) -> bool: ...

    async def files(self, name: Optional[str] = None) -> List[FormPart]: ...

    async def json(self, loads: Callable[[str], Any]=json.loads) -> Any: ...

    def has_body(self) -> bool: ...

    @property
    def charset(self) -> str: ...


Cookies = Dict[str, Cookie]


def method_without_body(method: str) -> bool: ...


class Request(Message):

    def __init__(self,
                 method: str,
                 url: bytes,
                 headers: Optional[List[HeaderType]]):
        self.method = ... # type: str
        self.url = ... # type: URL
        self.headers = headers
        self.route_values: Optional[Dict[str, str]] = ...
        self.content: Optional[Content] = ...
        self.identity: Union[None, Identity, User] = ...

    @classmethod
    def incoming(cls, method: str, path: bytes, query: bytes, headers: List[HeaderType]) -> Request: ...

    @property
    def query(self) -> Dict[str, List[str]]: ...

    @property
    def url(self) -> URL: ...

    @url.setter
    def url(self, value: Union[URL, bytes, str]): ...

    def __repr__(self):
        return f'<Request {self.method} {self.url.value.decode()}>'

    @property
    def cookies(self) -> Cookies: ...

    def get_cookies(self) -> Cookies: ...

    def get_cookie(self, name: bytes) -> Optional[Cookie]: ...

    def set_cookie(self, cookie: Cookie): ...

    def set_cookies(self, cookies: List[Cookie]): ...

    @property
    def etag(self) -> Optional[bytes]: ...

    @property
    def if_none_match(self) -> Optional[bytes]: ...

    def expect_100_continue(self) -> bool: ...


class Response(Message):

    def __init__(self,
                 status: int,
                 headers: Optional[List[HeaderType]] = None,
                 content: Optional[Content] = None):
        self.__headers = headers or []
        self.status = status
        self.content = content

    def __repr__(self):
        return f'<Response {self.status}>'

    @property
    def cookies(self) -> Cookies: ...

    @property
    def reason(self) -> str: ...

    def get_cookies(self) -> Cookies: ...

    def get_cookie(self, name: bytes) -> Optional[Cookie]: ...

    def set_cookie(self, cookie: Cookie): ...

    def set_cookies(self, cookies: List[Cookie]): ...

    def unset_cookie(self, name: bytes): ...

    def remove_cookie(self, name: bytes): ...

    def is_redirect(self) -> bool: ...



import inspect
from typing import Union, List, Sequence, Set, Tuple
from inspect import Signature, _empty
from .bindings import (FromJson,
                       FromQuery,
                       FromRoute,
                       FromServices,
                       ControllerBinder,
                       Binder,
                       FromBody,
                       RequestBinder,
                       ExactBinder,
                       IdentityBinder)
from guardpost.authentication import User, Identity
from blacksheep.normalization import copy_special_attributes


_next_handler_binder = object()


class NormalizationError(Exception):
    ...


class UnsupportedSignatureError(NormalizationError):

    def __init__(self, method):
        super().__init__(f'Cannot normalize method `{method.__qualname__}` because its signature contains '
                         f'*args or *kwargs parameters. If you use a decorator, please use `functools.@wraps` '
                         f'with your wrapper, to fix this error.')


class MultipleFromBodyBinders(NormalizationError):

    def __init__(self, method, first_match, new_match):
        super().__init__(f'Cannot use more than one `FromBody` binder for the same method ({method.__qualname__}). '
                         f'The first match was: {first_match}, a second one {new_match}.')


class AmbiguousMethodSignatureError(NormalizationError):

    def __init__(self, method):
        super().__init__(f'Cannot normalize method `{method.__qualname__}` due to its ambiguous signature. '
                         f'Please specify exact binders for its arguments.')


class RouteBinderMismatch(NormalizationError):

    def __init__(self, parameter_name, route):
        super().__init__(f'The parameter {parameter_name} for method {route.handler.__name__} is bound to route path, '
                         f'but the route doesn`t contain a parameter with matching name.')


def get_from_body_parameter(method) -> FromBody:
    """Extracts a single FromBody parameter from the given signature,
    throwing exception if more than one is defined."""
    sig = Signature.from_callable(method)

    from_body_parameter = None

    for name, parameter in sig.parameters.items():
        if isinstance(parameter.annotation, FromBody):
            if from_body_parameter is None:
                from_body_parameter = parameter.annotation
            else:
                raise MultipleFromBodyBinders(method, from_body_parameter, parameter)

    return from_body_parameter


_simple_types_handled_with_query = {
    str,
    int,
    float,
    bool,
    list,
    set,
    tuple,
    List[str],
    List[int],
    List[float],
    List[bool],
    Sequence[str],
    Sequence[int],
    Sequence[float],
    Sequence[bool],
    Set[str],
    Set[int],
    Set[float],
    Set[bool],
    Tuple[str],
    Tuple[int],
    Tuple[float],
    Tuple[bool],
}


def _check_union(parameter, annotation, method):
    """Checks if the given annotation is Optional[] - in such case unwraps it and returns its value

    An exception is thrown if other kinds of Union[] are used, since they are not supported by method normalization.
    In such case, the user of the library should read the desired value from the request object.
    """

    if hasattr(annotation, '__origin__') and annotation.__origin__ is Union:
        # support only Union[None, Type] - that is equivalent of Optional[Type]
        if type(None) not in annotation.__args__ or len(annotation.__args__) > 2:
            raise NormalizationError(f'Unsupported parameter type "{parameter.name}" for method "{method.__name__}"; '
                                     f'only Optional types are supported for automatic binding. '
                                     f'Read the desired value from the request itself.')

        for possible_type in annotation.__args__:
            if type(None) is possible_type:
                continue
            return True, possible_type
    return False, annotation


def get_parameter_binder(parameter, services, route, method):
    name = parameter.name

    if name == 'request':
        return RequestBinder()

    if name == 'services':
        return ExactBinder(services)

    original_annotation = parameter.annotation

    if original_annotation is _empty:
        if route:
            # 1. does route contain a parameter with matching name?
            if name in route.param_names:
                return FromRoute(str, name)

        # 2. do services contain a service with matching name?
        if name in services:
            return FromServices(name, services)

        # 3. does it use a specific name? (convention over configuration for common scenarios);
        if name == 'user' or name == 'identity':
            return IdentityBinder()

        # 4. default to query parameter
        return FromQuery(List[str], name)

    # unwrap the Optional[] annotation, if present:
    is_optional, annotation = _check_union(parameter, original_annotation, method)

    # 1. is the type annotation already a binder?
    if isinstance(annotation, Binder):

        if not annotation.name:
            # force name == parameter name
            annotation.name = name

        if route:
            if isinstance(annotation, FromRoute) and annotation.name not in route.param_names:
                raise RouteBinderMismatch(annotation.name, route)

        if isinstance(annotation, FromServices):
            annotation.services = services

        return annotation

    # 2A. do services contain a service with matching name?
    if name in services:
        return FromServices(name, services)

    # 2B. do services contain a service with matching type?
    if annotation in services:
        return FromServices(annotation, services)

    # 3. does route contain a parameter with matching name?
    if route and name in route.param_names:
        return FromRoute(annotation, name)

    # 4. is simple type?
    if annotation in _simple_types_handled_with_query:
        return FromQuery(annotation, name, required=not is_optional)

    if annotation is User or annotation is Identity:
        return IdentityBinder()

    # 5. from json body (last default)
    return FromJson(annotation, required=not is_optional)


def _get_binders_for_method(method, services, route):
    signature = Signature.from_callable(method)
    parameters = signature.parameters
    from_body_binder = None

    binders = []
    for parameter_name, parameter in parameters.items():
        if not route and parameter_name in {'handler', 'next_handler'}:
            binders.append(_next_handler_binder)
            continue

        binder = get_parameter_binder(parameter, services, route, method)
        if isinstance(binder, FromBody):
            if from_body_binder is None:
                from_body_binder = binder
            else:
                raise AmbiguousMethodSignatureError(method)
        binders.append(binder)
    return binders


def get_binders(route, services):
    """Returns a list of binders to extract parameters for a request handler."""
    return _get_binders_for_method(route.handler, services, route)


def get_binders_for_middleware(method, services):
    return _get_binders_for_method(method, services, None)


def _copy_name_and_docstring(source_method, wrapper):
    try:
        wrapper.__name__ = source_method.__name__
        wrapper.__doc__ = source_method.__doc__
    except AttributeError:
        pass


def _get_sync_wrapper_for_controller(binders, method):
    async def handler(request):
        values = []
        controller = await binders[0].get_value(request)
        await controller.on_request(request)

        values.append(controller)

        for binder in binders[1:]:
            values.append(await binder.get_value(request))

        response = method(*values)
        await controller.on_response(response)
        return response
    return handler


def _get_async_wrapper_for_controller(binders, method):
    async def handler(request):
        values = []
        controller = await binders[0].get_value(request)
        await controller.on_request(request)

        values.append(controller)

        for binder in binders[1:]:
            values.append(await binder.get_value(request))

        response = await method(*values)
        await controller.on_response(response)
        return response

    return handler


def get_sync_wrapper(services, route, method, params, params_len):
    if params_len == 0:
        # the user defined a synchronous request handler with no input
        async def handler(_):
            return method()

        return handler

    if params_len == 1 and 'request' in params:
        async def handler(request):
            return method(request)

        return handler

    binders = get_binders(route, services)

    if isinstance(binders[0], ControllerBinder):
        return _get_sync_wrapper_for_controller(binders, method)

    async def handler(request):
        values = []
        for binder in binders:
            values.append(await binder.get_value(request))
        return method(*values)

    return handler


def get_async_wrapper(services, route, method, params, params_len):
    if params_len == 0:
        # the user defined a request handler with no input
        async def handler(_):
            return await method()

        return handler

    if params_len == 1 and 'request' in params:
        # no need to wrap the request handler
        return method

    binders = get_binders(route, services)

    if isinstance(binders[0], ControllerBinder):
        return _get_async_wrapper_for_controller(binders, method)

    async def handler(request):
        values = []
        for binder in binders:
            values.append(await binder.get_value(request))
        return await method(*values)

    return handler


def normalize_handler(route, services):
    method = route.handler

    sig = Signature.from_callable(method)
    params = sig.parameters
    params_len = len(params)

    if any(str(param).startswith('*') for param in params.values()):
        raise UnsupportedSignatureError(method)

    if inspect.iscoroutinefunction(method):
        normalized = get_async_wrapper(services, route, method, params, params_len)
    else:
        normalized = get_sync_wrapper(services, route, method, params, params_len)

    if normalized is not method:
        copy_special_attributes(method, normalized)
        _copy_name_and_docstring(method, normalized)

    return normalized


def _is_basic_middleware_signature(parameters):
    values = list(parameters.values())

    if len(values) != 2:
        return False

    first_one = values[0]
    second_one = values[1]
    if first_one.name == 'request' and second_one.name in {'handler', 'next_handler'}:
        return True
    return False


def _get_middleware_async_binder(method, services):
    binders = get_binders_for_middleware(method, services)

    async def handler(request, next_handler):
        values = []
        for binder in binders:
            if binder is _next_handler_binder:
                values.append(next_handler)
            else:
                values.append(await binder.get_value(request))

        if _next_handler_binder in binders:
            # middleware that can continue the chain: control is left to it;
            # for example an authorization middleware can decide to now call the next handler
            return await method(*values)

        # middleware that cannot continue the chain, so we continue it here
        await method(*values)
        return await next_handler(request)

    return handler


def normalize_middleware(middleware, services):
    sig = Signature.from_callable(middleware)
    params = sig.parameters

    if _is_basic_middleware_signature(params):
        return middleware

    if inspect.iscoroutinefunction(middleware):
        normalized = _get_middleware_async_binder(middleware, services)
    else:
        raise ValueError('Middlewares must be asynchronous functions')

    return normalized



import os
import logging
import logging.handlers
from datetime import datetime
from blacksheep.utils.folders import ensure_folder
from blacksheep.exceptions import HttpException, MessageAborted


def _get_loggers():
    process_id = os.getpid()

    access_logger = logging.getLogger('blacksheep.server-access')
    app_logger = logging.getLogger('blacksheep.server')

    app_logger.setLevel(logging.INFO)
    access_logger.setLevel(logging.DEBUG)

    max_bytes = 24 * 1024 * 1024

    file_handler = logging.handlers.RotatingFileHandler

    now = datetime.now()
    ts = now.strftime('%Y%m%d')
    hour_ts = now.strftime('%H%M%S')

    ensure_folder(f'logs/{ts}')

    access_handler = file_handler(f'logs/{ts}/{hour_ts}-blacksheep-access-{process_id}.log',
                                  maxBytes=max_bytes,
                                  backupCount=5)

    app_handler = file_handler(f'logs/{ts}/{hour_ts}-blacksheep-{process_id}.log',
                               maxBytes=max_bytes,
                               backupCount=5)

    access_handler.setLevel(logging.DEBUG)
    app_handler.setLevel(logging.DEBUG)
    app_handler.setFormatter(logging.Formatter(
        '%(levelname)s @ %(asctime)s @ %(filename)s '
        '%(funcName)s %(lineno)d: %(message)s'))

    access_logger.addHandler(access_handler)
    app_logger.addHandler(app_handler)

    access_handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(message)s'))

    return access_logger, app_logger


def get_logged_url(request):
    if request.url.query:
        return request.url.path.decode() + '?<query is hidden>'
    return request.url.path.decode()


def setup_sync_logging():
    access_logger, app_logger = _get_loggers()

    async def logging_middleware(request, handler):
        access_logger.debug(f'{request.method.ljust(8)} {get_logged_url(request)}')
        try:
            response = await handler(request)
        except HttpException:
            raise
        except MessageAborted:
            app_logger.warning(f'The connection was lost or aborted while the request was being sent. '
                               f'{request.method.ljust(8)} {get_logged_url(request)}')
            raise
        except Exception:
            app_logger.exception(f'{"*" * 30}\nunhandled exception while handling: {request.method} {get_logged_url(request)}')
            raise
        return response

    return logging_middleware, access_logger, app_logger




from typing import Union, Optional


class InvalidURL(Exception):

    def __init__(self, message: str): ...


class URL:

    def __init__(self, value: bytes):
        self.value = ... # type: bytes
        self.schema = ... # type: Optional[bytes]
        self.host = ... # type: Optional[bytes]
        self.port = ... # type: int
        self.path = ... # type: bytes
        self.query = ... # type: bytes
        self.fragment = ... # type: Optional[bytes]
        self.is_absolute = ... # type: bool

    def __repr__(self):
        return f'<URL {self.value}>'

    def join(self, other: URL) -> URL: ...

    def base_url(self) -> URL: ...

    def __add__(self, other: Union[bytes, URL]): ...

    def __eq__(self, other: URL) -> bool: ...



from blacksheep import Request, Response
from blacksheep.exceptions import HttpException
from blacksheep.server.routing import Router
from typing import Dict, Union, Type, Callable


ExceptionHandlersType = Dict[Union[int, Type[Exception]], Callable]


class BaseApplication:

    def __init__(self, show_error_details: bool, router: Router):
        self.router = router
        self.exceptions_handlers = self.init_exceptions_handlers()
        self.show_error_details = show_error_details

    def init_exceptions_handlers(self) -> ExceptionHandlersType: ...

    async def handle(self, request: Request) -> Response: ...

    async def handle_internal_server_error(self, request: Request, exc: Exception) -> Response: ...

    async def handle_http_exception(self, request: Request, http_exception: HttpException) -> Response: ...

    async def handle_exception(self, request: Request, exc: Exception) -> Response: ...



import re
from typing import Union


BytesOrStr = Union[bytes, str]


def ensure_bytes(value: BytesOrStr):
    if isinstance(value, bytes):
        return value
    if isinstance(value, str):
        return value.encode()
    raise ValueError('Expected bytes or str')


def ensure_str(value: BytesOrStr):
    if isinstance(value, str):
        return value
    if isinstance(value, bytes):
        return value.decode()
    raise ValueError('Expected bytes or str')


def remove_duplicate_slashes(value: str) -> str:
    return re.sub('/{2,}', '/', value)


def join_fragments(*args: BytesOrStr) -> str:
    """Joins URL fragments bytes"""
    return '/' + '/'.join(remove_duplicate_slashes(ensure_str(arg)).strip('/') for arg in args)



import os
import uuid
import logging.handlers
from datetime import datetime
from blacksheep.utils.folders import ensure_folder
from .exceptions import InvalidResponseException


def _get_logger():
    process_id = os.getpid()
    
    client_logger = logging.getLogger('blacksheep.client')

    if __debug__:
        client_logger.setLevel(logging.DEBUG)
    else:
        client_logger.setLevel(logging.INFO)

    now = datetime.now()
    ts = now.strftime('%Y%m%d')
    hour_ts = now.strftime('%H%M%S')

    ensure_folder(f'logs/{ts}')

    max_bytes = 24 * 1024 * 1024

    file_handler = logging.handlers.RotatingFileHandler

    client_handler = file_handler(f'logs/{ts}/{hour_ts}-blacksheep-client-{process_id}.log',
                                  maxBytes=max_bytes,
                                  backupCount=5)
    client_handler.setLevel(logging.DEBUG)
    client_handler.setFormatter(logging.Formatter('%(levelname)s %(asctime)s @ BlackSheep.Client: %(message)s'))

    client_logger.addHandler(client_handler)

    return client_logger


def get_logged_url(request):
    if b'?' in request.url.value:
        return request.url.path.decode() + '?<query is hidden>'
    return request.url.path.decode()


def get_response_record(response, trace_id, method, logged_url):
    message = f'({trace_id}) RECEIVED: {str(response.status).ljust(8)} FOR {method} {logged_url}'
    if response.is_redirect():
        location = response.headers.get_single(b'location')
        if location:
            return message + ' REDIRECT > ' + location.value.decode()
    return message


def get_client_logging_middleware():
    client_logger = _get_logger()

    async def client_logging_middleware(request, handler):
        trace_id = uuid.uuid4()
        logged_url = get_logged_url(request)

        request.trace_id = trace_id
        client_logger.debug(f'({trace_id}) SENDING : {request.method.ljust(8)} {logged_url}')
        try:
            response = await handler(request)
            client_logger.debug(get_response_record(response, trace_id, request.method, logged_url))
        except InvalidResponseException as invalid_response_ex:
            client_logger.warning(f'({trace_id}) SERVER PRODUCED AN INVALID RESPONSE FOR: {request.method} '
                                f'{get_logged_url(request)} - {invalid_response_ex}')
            raise
        except TimeoutError as timeout:
            client_logger.warning(f'({trace_id}) OPERATION TIMEOUT: {request.method} '
                                f'{get_logged_url(request)} - {timeout}')
            raise
        except Exception:
            client_logger.exception(f'({trace_id}) UNHANDLED EXCEPTION WHILE HANDLING: {request.method} '
                                    f'{get_logged_url(request)}')
            raise
        return response

    return client_logging_middleware



from typing import Optional
from datetime import datetime
from urllib.parse import quote, unquote


cpdef bytes datetime_to_cookie_format(object value):
    return value.strftime('%a, %d %b %Y %H:%M:%S GMT').encode()


cpdef object datetime_from_cookie_format(bytes value):
    value_str = value.decode()
    try:
        return datetime.strptime(value_str, '%a, %d %b %Y %H:%M:%S GMT')
    except ValueError:
        return datetime.strptime(value_str, '%a, %d-%b-%Y %H:%M:%S GMT')


cdef class Cookie:

    def __init__(self,
                 bytes name,
                 bytes value,
                 bytes expires=None,
                 bytes domain=None,
                 bytes path=None,
                 bint http_only=0,
                 bint secure=0,
                 bytes max_age=None,
                 bytes same_site=None):
        if not name:
            raise ValueError('A cookie name is required')
        self.name = name
        self.value = value
        self.expires = expires
        self._expiration = None
        self.domain = domain
        self.path = path
        self.http_only = http_only
        self.secure = secure
        self.max_age = max_age
        self.same_site = same_site

    cpdef Cookie clone(self):
        return Cookie(self.name,
                      self.value,
                      self.expires,
                      self.domain,
                      self.path,
                      self.http_only,
                      self.secure,
                      self.max_age,
                      self.same_site)

    @property
    def expiration(self):
        if not self.expires:
            return None

        if self._expiration is None:
            self._expiration = datetime_from_cookie_format(self.expires)
        return self._expiration

    @expiration.setter
    def expiration(self, value):
        self._expiration = value
        if value:
            self.expires = datetime_to_cookie_format(value)
        else:
            self.expires = None

    cpdef void set_max_age(self, int max_age):
        self.max_age = str(max_age).encode()

    def __eq__(self, other):
        if isinstance(other, str):
            return other.encode() == self.value
        if isinstance(other, bytes):
            return other == self.value
        if isinstance(other, Cookie):
            return other.name == self.name and other.value == self.value
        return NotImplemented

    def __repr__(self):
        return f'<Cookie {self.name.decode()}: {self.value.decode()}>'


cpdef Cookie parse_cookie(bytes value):
    cdef bytes eq, expires, domain, path, part, max_age, k, v, lower_k, lower_part
    cdef bint http_only, secure
    cdef bytes same_site
    cdef list parts
    eq = b'='
    parts = value.split(b'; ')
    if len(parts) == 0:
        try:
            name, value = value.split(eq)
        except ValueError as unpack_error:
            raise ValueError(f'Invalid name=value fragment: {parts[0]}')
        else:
            return Cookie(name, value)
    if len(parts) == 1:
        # some set a cookie with a separator without space
        parts = value.split(b';')
    try:
        name, value = parts[0].split(eq)
    except ValueError as unpack_error:
        raise ValueError(f'Invalid name=value fragment: {parts[0]}')

    if b' ' in value and value.startswith(b'"'):
        value = value.strip(b'"')

    expires = None
    domain = None
    path = None
    http_only = False
    secure = False
    max_age = None
    same_site = None

    for part in parts:
        if eq in part:
            k, v = part.split(eq)
            lower_k = k.lower()
            if lower_k == b'expires':
                expires = v
            elif lower_k == b'domain':
                domain = v
            elif lower_k == b'path':
                path = v
            elif lower_k == b'max-age':
                max_age = v
            elif lower_k == b'samesite':
                same_site = v
        else:
            lower_part = part.lower()
            if lower_part == b'httponly':
                http_only = True
            if lower_part == b'secure':
                secure = True

    return Cookie(unquote(name.decode()).encode(),
                  unquote(value.decode()).encode(),
                  expires,
                  domain,
                  path,
                  http_only,
                  secure,
                  max_age,
                  same_site)


cdef bytes write_cookie_for_response(Cookie cookie):
    cdef list parts = []
    parts.append(quote(cookie.name).encode() + b'=' + quote(cookie.value).encode())

    if cookie.expires:
        parts.append(b'Expires=' + cookie.expires)

    if cookie.max_age:
        parts.append(b'Max-Age=' + cookie.max_age)

    if cookie.domain:
        parts.append(b'Domain=' + cookie.domain)

    if cookie.path:
        parts.append(b'Path=' + cookie.path)

    if cookie.http_only:
        parts.append(b'HttpOnly')

    if cookie.secure:
        parts.append(b'Secure')

    if cookie.same_site:
        parts.append(b'SameSite=' + cookie.same_site)

    return b'; '.join(parts)
